/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*\\\\\\\\\\\\\\\\\\\\\\\\\\  DRAGDROP.C    \\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*                                                                      */
/*    Routine pour Drag and drop sous MultiTos                          */
/*    source: D&D Atari, demo OEP (Alexander Lorenz)                    */
/*    remise en page traduction pour adaptation EZ_GEM: Roland SEVIN    */
/*    Donnee Globale necessaire:  Appl_id identificateur de             */
/*                                        l'application declare dans    */
/*    (Tab = 4)                           EAZY_GEM.H inclus  cet effet */
/*                                                                      */
#include "EAZY_GEM.H"  /* inclure les definitions des fonctions que le  */
                       /* que le module EAZY_GEM.C met  notre          */
                       /* notre disposition                             */
                       /* et inclure les definitions des variables      */
                       /* globales  EAZY_GEM                            */

/* end of include....................................................*/
/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*\\\\\\\\\\\\\\ V A R I A B L E S    G L O B A L E S \\\\\\\\\\\\\\\\\\*/
/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

/*................ nom du fichier pipe de D&D...........................*/
static BYTE Pipename[] = "U:\\PIPE\\DRAGDROP.AA"; 


/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*\\\\\\\\\\\\\\\ P R O T O T Y P E S  F O N C T I O N S \\\\\\\\\\\\\\\*/
/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*------------------------ PutDragDropData -----------------------------*/
/* ACTION: place des donnees sur le pipe Drag & Drop. Cette fonction    */
/*         est  appeler apres relachage du bouton de la  souris au     */
/*         dessus d'une fenetre cliente, pour luis fournir des donnes  */
/*         (  la condition qu'elle accepte ces donnes)                */
/*         elle recherche la fenetre cliente, et essaie de transmettre  */
/*         par le protocole DRAG & DROP les donnes.                    */
/* ENTREE: char *data: pointe sur la zone des donnes.                  */
/*         long len_data: longueur des donnes a transmettre            */
/*         long must_accept_type: type de donnee: 'ARGS'  '.TXT'        */
/*                                '.IMG' etc....                        */
/*         INT16 x, INT16 y: position de la souris lors du lacher bouton    */
/*         INT16 ksp: etat des touches SHIFTn, CTRL, ALT, CAPS lors du    */
/*                  lacher de bouton de souris.                         */
/* RETOUR: 0/si echec    longueur des donnees transmises/si OK          */

long PutDragDropData(char *data, long len_data, long must_accept_type, INT16 x, INT16 y, INT16 ksp);

/*------------------- GetDragDropData -------------------------------------*/
/* ACTION: recupere les donnes presentes sur le pipe du drag & drop       */
/*         lors du message AP_DRAGDROP.                                    */
/* NOTE:   cette fonction est polymorphe, Si le dst d'entree (dst) ou      */
/*         la longueur de recopie (len_dst) sont egals a zero, il est      */
/*         retourne en ret l'adresse d'un buffer dans lequel se trouve les */
/*         donnes (termine par zero de fin)                              */
/*         Il sera a votre charge de liberer ce buffer avec la fonction:   */
/*         free();                                                         */
/* ENTREE: dd_ext: extension du fichier Drag & Drop tel qu'il est fourni   */
/*                 dans le mot N7 lors reception du message  APP_DRAGDROP */
/*                 Exemple: Evnt->pipe[7]                                  */
/*         char *dst: pointeur sur la chaine dans laquelle le texte sera   */
/*               recopie. (Si zero alors le retour est l'adresse d'un     */
/*                          BUFFER dans lequel les donnees seront placees  */
/*                          Apres usage il vous faudra liberer ce buffer   */
/*                          par free() ).                                  */
/*         long len_dst: longueur de la chaine d'entree, afin que la       */
/*               recopie ne deborde pas.                                   */
/*         long *recept_type: reoit en sortie le type de donnes que le   */
/*                            drag & drop nous envoie. Si zero alors echec */
/*                            de la recuperation.                          */
/*         char *acceptDataTyp: pointeur sur une zone memoire de n fois    */
/*                              4 octets representant le type de           */
/*                              donnes que l'on accepte. Exemple:         */
/*                              "ARGS.EXT.ASC"                             */
/*                              maximum 8 arguments soit 32 octets.        */
/* RETOUR: long: longueur de la chaine pouvant etre recupre. Si cette    */
/*              longueur est plus longue que len_dst, c'est qu'il y a eu   */
/*              non recopie d'une partie de la chaine.                     */
/*              Si dst ou len_dst sont egals  zero, ret est l'adresse     */
/*              du buffer reserve par la fonction pour stocker les donnees */
/*              recuperees dans le DRAG & DROP. Dans ce cas il vous faudra */
/*              liberer ce buffer apres usage avec la fonction free().     */
/*              les 4 premiers octets de ce buffer sont alors la longueur  */
/*              des donnes qui suivent. les donnes sont toujours fermes */
/*              par un zero de fin                                         */

long GetDragDropData(INT16 dd_ext, char *dst, long len_dst, long *recept_type, char *acceptDataTyp);

/*--------------------------- DragDropCreate -----------------------------*/
/* ACTION: Routine pour l'envoyeur qui cree un pipe D&D                   */
/* ENTREE: WORD app_id: Iddentificateur d'application de l'expediteur     */
/*         WORD rcvr_id: Iddentificateur d'application du destinataire    */
/*         WORD window: handle GEM de la fenetre destinataire             */
/*         WORD mx, WORD my: coordonnes souris (lors lacher du D&D).     */
/*         WORD kbstate: etat touches controles du clavier(lors lacher    */
/*                       du D&D)                                          */
/*         ULONG format[8]; pointe sur un tableau de 8 longs qui recevra  */
/*                          les formats acceptes par le destinataire.     */
/*         void **oldpipesig: pointeur sur pointeur de void.              */
/* RETOUR: un nombre negatif ou = 0 si echec -1 echec -2 appl_write:      */
/*         le handle du pipe si tout est OK avec l'identificateur         */
/*                   ou extension dans le mot pointe par WORD *pipe       */

WORD   DragDropCreate( WORD   app_id, WORD rcvr_id, WORD window, WORD mx, WORD my, 
                WORD kbstate, ULONG format[8], void **oldpipesig );
                
/*--------------------- DragDropOpen ----------------------------------*/
/* ACTION: ouvre un pipe deja cree par un expediteur, en reponse au    */
/*         au message AP_DRAGDROP.                                     */
/* ENTREE: WORD ext_name: contient en poids fort et poids faible les   */
/*                        deux caracteres d'extension du pipe          */
/*                    reu par le destinataire lors reception du       */
/*                    du message: AP_DRAGDROP, dans le mot N7:        */
/*         ULONG format[8]: pointeur sur les formats supportes         */
/*         void **oldpipesig: pointeur sur pointeur de void.           */
/* RETOUR:  >0 handle du pipe.                                         */
/*         <=0 erreur                                                  */

WORD DragDropOpen( WORD ext_name, ULONG format[8], void **oldpipesig );

/*--------------------------- ddcreate -----------------------------*/
/* ACTION: Routine pour l'envoyeur qui cree un pipe D&D             */
/* ENTREE: WORD *pipe: pointe sur un INT16 qui recevra: dans le poids */
/*                     fort et le faible l'extension du pipe cree.  */
/*                     Cette extention valide est automatiquement   */
/*                     recherche par la routine.                   */
/*         void **oldpipesig: pointeur sur pointeur de void.        */
/* RETOUR: un nombre negatif ou = 0 si echec:                       */
/*         le handle du pipe si tout est OK avec l'identificateur   */
/*                   ou extension dans le mot pointe par WORD *pipe */

WORD ddcreate(WORD *pipe, void **oldpipesig);

/*--------------------- ddmessage -------------------------------*/
/* ACTION: envoie un message D&D: AP_DRAGDROP  une application  */
/* ENTREE: WORD apid: ID application destinataire du message     */
/*         WORD fd :  handle du Pipe c'est pas l'extension gaffe */
/*                    mais celui retourne par ddcreate           */
/*         WORD winid: handle GEM de la fenetre cible.           */
/*                     obtenu en faisant un wind_find()          */
/*         WORD mx, WORD my, WORD kstate: etat souris et touches */
/*                  de controles clavier.                        */
/*         WORD Pipename: extension du pipe (deux dernier octets */
/*                        de Pipename[]:  Pipename[17] & [18]    */
/*                        ou le mot: WORD *pipe initialise par   */
/*                        la fonction: ddcreate()                */
/*         void *oldpipesig: pointeur sur void.                  */
/* RETOUR: >0 pas d'erreur                                       */
/*        -1 DD_NAK: (non accept)  alors le pipe est ferme.     */
/*        -2 Timeout pas de reponse  alors le pipe est ferme.    */
/*        -3: erreur par appl_write()  alors le pipe est ferme.  */

WORD ddmessage(WORD apid, WORD fd, WORD winid, WORD mx, WORD my, WORD kstate, 
               WORD Pipename, void *oldpipesig);

/*------------------- ddrexts -----------------------------------*/
/* ACTION: Demande au destinataire la liste des types de donnees */
/*         qu'il peut accepter. liste de 32 long ou byte[4]      */
/* ENTREE: WORD fd :  handle du Pipe c'est pas l'extension gaffe */
/*                    mais celui retourne par ddcreate           */
/*         BYTE *exts: pointe sur zone de 32 octets ou sera      */
/*                     retourne les extensions admises.         */
/*         void *oldpipesig: pointeur  de void.                  */
/* RETOUR: WORD :  >0 pas d'erreur.                              */
/*                <=0 0 alors le pipe est ferme.                 */

WORD ddrexts(WORD fd, BYTE *exts,  void *oldpipesig);

/*------------------------- ddsexts --------------------------------------*/
/* ACTION: envoie a l'expediteur la liste des types de donnes acceptes  */
/*         cette liste est donc de 32 octets, chaque type etant 4 octets  */
/*         du type "ARGS" ".IMG" etc..  elle peut accepter 8 types de     */
/*         donnes. Si elle n'est pas complete (exemple seulement deux    */
/*         types de donnes) le reste doit etre rempli de zeros.          */
/* ENTREE: BYTE *data_typ: pointe sur liste type de donnees acceptes     */
/*                         "ARGS" "PATH" ".IMG" ".TXT" etc...             */
/*         void *oldpipesig: pointeur sur  void.                          */
/* RETOUR:  >0 tout est OK                                                */
/*        <= 0 erreur d'ecriture (le pipe est alors ferme)                */

WORD ddsexts(WORD fd, BYTE *data_typ, void *oldpipesig);

/*------------------- ddstry ---------------------------------------*/
/* ACTION: demande au destinataire si il accepte un certain         */
/*         type de donnees dont il presente les caracteristiques    */
/*         data_typ, LONG size etc ....                             */
/* ENTREE: WORD fd :  handle du Pipe c'est pas l'extension gaffe    */
/*                    mais celui retourne par ddcreate              */
/*         BYTE *data_typ: pointe sur type de donnees a envoyer     */
/*                         "ARGS" "PATH" ".IMG" ".TXT" etc...       */
/*              BYTE *text: descripteur des donnes exemple:        */
/*                       "Caractere ASCII" est optionnel on peut    */
/*                        passer une chaine vide.                   */
/*              BYTE *name: non de fichier associe  la donne      */
/*                        ce parametre est aussi optionnel et       */
/*                        peut etre une chaine vide.                */
/*              LONG size: longueur des donnes  transmettre       */
/* RETOUR: un INT16:                                                  */
/*       DD_OK  le destinataire ou cible accepte les donnes        */
/*       DD_NAK le D&D doit etre interrompu                         */
/*       DD_EXT le cible refuse les donnes telles quelles          */
/*       DD_LEN le cible ne peut accepter la longueur               */
/*     DD_TRASH le cible repond que donnes sont sur une poubelle   */
/*   DD_PRINTER le cible repond que donnes sont sur une Imprimante */
/* DD_CLIPBOARD le cible repond que donnes sont sur le CLIPBOARD   */

WORD ddstry(WORD fd, BYTE *data_typ, BYTE *text, BYTE *name, LONG size);

/*-------------------------- ddclose --------------------------------*/
/* ACTION: ferme le pipe et met fin aux procedures d'echanges        */
/*         peut etre utilisee par l'expediteur et le destinataire.   */
/* ENTREE: WORD fd :  handle du Pipe c'est pas l'extension gaffe     */
/*                    mais celui retourne par ddcreate               */
/*         void *oldpipesig:  pointeur de void.                      */

VOID ddclose(WORD fd, void *oldpipesig );

/*--------------------- ddopen ----------------------------------------*/
/* ACTION: ouvre un pipe deja cree par un expediteur, en reponse au    */
/*         au message AP_DRAGDROP.                                     */
/* ENTREE: WORD ddnam: extension du pipe  ouvrir tel qu'il a ete recu */
/*                     par le destinataire lors reception du message:  */
/*                     AP_DRAGDROP, dans le mot N7: exemple pipe[7]   */
/*         BYTE ddmsg: DD_OK on accepte   DD_NAK on refuse             */
/*         void **oldpipesig: pointeur sur pointeur de void.           */
/* RETOUR:  >0 handle du pipe.                                         */
/*         <=0 erreur                                                  */
WORD ddopen(WORD ddnam, BYTE ddmsg, void **oldpipesig);

/*------------------------- ddrtry -----------------------------------*/
/* ACTION: demande  l'expediteur, en lui precisant le type de donnee */
/*         que l'on accepte de recevoir, des renseignements divers    */
/*         dessus notamment la longueur.                              */
/* ENTREE: WORD fd :  handle du Pipe c'est pas l'extension gaffe      */
/*                    mais celui retourne par ddcreate                */
/*         BYTE *whichext: pointe sur type de donnees que l'on        */
/*                         souhaite recevoir. zone de 5 octets.       */
/*                         "ARGS" "PATH" ".IMG" ".TXT" etc...         */
/* SORTIE:                                                            */
/*         BYTE *whichext: est rempli apre appel avec type de donnees */
/*                         reues "ARGS" "PATH" ".IMG" ".TXT" etc...  */
/*                         si different de celui en entree alors      */
/*                         erreur.                                    */
/*         BYTE *text: descripteur des donnes exemple:               */
/*                    "Caractere ASCII" initialisee par la fonction   */
/*                    la longueur de la zone pointee par text doit    */
/*                    etre au minimum de: DD_NAMEMAX 128 octets       */
/*         BYTE *name: non de fichier associe  la donne             */
/*                     ce parametre est initialise par la fonction    */
/*                     la longueur de la zone pointee par name doit   */
/*                     etre au minimum de: DD_NAMEMAX 128 octets      */
/*         LONG *size: pointe sur longueur des donnes  ecevoir      */
/*                     cette valeur est initialise en retour.        */
/* RETOUR: >0 Tout est OK                                             */
/*         -1 erreur                                                  */

WORD ddrtry(WORD fd, BYTE *name, BYTE *file, BYTE *whichext, LONG *size);

/*-------------------- ddreply ----------------------------------------*/
/* ACTION: envoie 1 octet exemple: DD_OK ou DD_NAK au pipe.            */
/* ENTREE: WORD fd :  handle du Pipe c'est pas l'extension gaffe       */
/*                    mais celui retourne par ddcreate ou ddopen       */
/*         BYTE ack: octet  envoyer                                   */
/* RETOUR: >=0 Tout est OK                                             */
/*           0 erreur                                                  */

WORD ddreply(WORD fd, BYTE ack);

/*------------------------- Parse_ARGS --------------------------------*/
/* ACTION: transforme une liste d'arguments soit separes par des       */
/*         espaces, en une suite de chaines terminees par des zeros de */
/*         fin. Si les arguments sont entre quotes: exemple            */
/*         'arguments1''argument2'  etc... cela marche aussi et les    */
/*         espaces ne seront pas transformes en zero de fin.           */
/* ENTREE: char *str: chaine d'arguments termine par un zero.         */
/* RETOUR: nombre d'arguments.                                         */

INT16 Parse_ARGS(char *str);

/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*                A   C O M M E N C E   I C I   !!!!                   */
/*       Definition et implementation des fonctions de ce module        */
/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

/*------------------------ PutDragDropData -----------------------------*/
/* ACTION: place des donnees sur le pipe Drag & Drop. Cette fonction    */
/*         est  appeler apres relachage du bouton de la  souris au     */
/*         dessus d'une fenetre cliente, pour lui fournir des donnes   */
/*         (  la condition qu'elle accepte ces donnes)                */
/*         elle recherche la fenetre cliente, et essaie de transmettre  */
/*         par le protocole DRAG & DROP les donnes.                    */
/* ENTREE: char *data: pointe sur la zone des donnes.                  */
/*         long len_data: longueur des donnes a transmettre            */
/*         long must_accept_type: type de donnee: 'ARGS'  '.TXT'        */
/*                                '.IMG' etc....                        */
/*         INT16 x, INT16 y: position de la souris lors du lacher bouton*/
/*         INT16 ksp: etat des touches SHIFTn, CTRL, ALT, CAPS lors du  */
/*                  lacher de bouton de souris.                         */
/* RETOUR: long: 0/si echec    longueur des donnees transmises/si OK    */

long PutDragDropData(char *data, long len_data, long must_accept_type, INT16 x, INT16 y, INT16 ksp)
{INT16 fd , hwnd, targID, ret;
 char  ext_type[32];
 char info[DD_NAMEMAX], name[ DD_NAMEMAX];
 void *oldpipesig       = 0;
 /*
 INT16 accept             = 0;
 long reply_accept_type = 0;
 INT16 i;
 */

 len_data          = strlen(data)+1;
 /*................. donnees a accepter .....................*/
 strnset( ext_type, '\0', 32);
 memcpy(  ext_type, &must_accept_type,4);/*"ARGS.TXT.ASCTEST");*/
 /*................. si pas de multi os cassos ..............*/
 if (MachineInfo.mult_os != 'MagX' && 
     MachineInfo.mult_os != 'MiNT'
    )      /* si Magic: MagX  Si Mint: MiNT  */                    return 0;
 /*.......... quelle est la fenetre sous souris ? ...........*/
 hwnd = wind_find(x, y);  if (hwnd==0)                             return 0;
 /*.......... a quelle application elle appartient ..........*/
 wind_get(hwnd, WF_OWNER, &targID, &fd, &fd, &fd);
 /*............. indispensable sous peine de non reponse ....*/
 wind_update(END_UPDATE);
 /*....... creer un pipe de DRAGDROP ........................*/
 fd = DragDropCreate( Appl_id, targID, hwnd, x, y, ksp, (ULONG*)ext_type, &oldpipesig);
 if (fd<= 0 )                                                      return 0;
 /*.............. voir si dans liste de types de donnes ....*/
 /*               acceptes par le destinataire              */
 /*               se trouve notre type                       */
 /*   NE FONCTIONNE PAS AVEC LUNA et GCCSHEL qui envoient    */
 /*   une liste vide. Donc ne pas verifier                   */
 /*
 for (i=0; i<8 && accept==0; i++)
     {memcpy(&reply_accept_type, ext_type + i*4, 4);
      if (reply_accept_type==must_accept_type) 
         {accept = 1;
          break;
         }
     }
 if (accept==0)  {ddclose(fd, oldpipesig);                         return 0;}
 */
 /*....... envoyer nos spec. de donnes au destinataire ......*/

 strcpy(  info , "That the crazy man");
 strcpy(  name , "");
 ret = ddstry( fd, (BYTE*)&must_accept_type, info, name, len_data);
 if (ret!=DD_OK)  { ddclose(fd, oldpipesig);                       return 0;}
 /*......... ecrire les donnes ..............................*/
 if (Fwrite(fd, len_data, data) != len_data)
    {ddclose(fd, oldpipesig);                                      return 0;
    }
 ddclose(fd, oldpipesig);
 return len_data;
}

/*------------------- GetDragDropData -------------------------------------*/
/* ACTION: recupere les donnes presentes sur le pipe du drag & drop       */
/*         lors du message AP_DRAGDROP.                                    */
/* NOTE:   cette fonction est polymorphe, Si le dst d'entree (dst) ou      */
/*         la longueur de recopie (len_dst) sont egals a zero, il est      */
/*         retourne en ret l'adresse d'un buffer dans lequel se trouve les */
/*         donnes (termine par zero de fin)                              */
/*         Il sera a votre charge de liberer ce buffer avec la fonction:   */
/*         free();                                                         */
/* ENTREE: dd_ext: extension du fichier Drag & Drop tel qu'il est fourni   */
/*                 dans le mot N7 lors reception du message  APP_DRAGDROP */
/*                 Exemple: Evnt->pipe[7]                                  */
/*         char *dst: pointeur sur la chaine dans laquelle le texte sera   */
/*               recopie. (Si zero alors le retour est l'adresse d'un     */
/*                          BUFFER dans lequel les donnees seront placees  */
/*                          Apres usage il vous faudra liberer ce buffer   */
/*                          par free() ).                                  */
/*         long len_dst: longueur de la chaine d'entree, afin que la       */
/*               recopie ne deborde pas.                                   */
/*         long *recept_type: reoit en sortie le type de donnes que le   */
/*                            drag & drop nous envoie. Si zero alors echec */
/*                            de la recuperation.                          */
/*         char *acceptDataTyp: pointeur sur une zone memoire de n fois    */
/*                              4 octets representant le type de           */
/*                              donnes que l'on accepte. Exemple:         */
/*                              "ARGS.EXT.ASC"                             */
/*                              maximum 8 arguments soit 32 octets.        */
/* RETOUR: long: longueur de la chaine pouvant etre recupre. Si cette    */
/*              longueur est plus longue que len_dst, c'est qu'il y a eu   */
/*              non recopie d'une partie de la chaine.                     */
/*              Si dst ou len_dst sont egals  zero, ret est l'adresse     */
/*              du buffer reserve par la fonction pour stocker les donnees */
/*              recuperees dans le DRAG & DROP. Dans ce cas il vous faudra */
/*              liberer ce buffer apres usage avec la fonction free().     */
/*              les 4 premiers octets de ce buffer sont alors la longueur  */
/*              des donnes qui suivent. les donnes sont toujours fermes */
/*              par un zero de fin                                         */

long GetDragDropData(INT16 dd_ext, char *dst, long len_dst, long *recept_type, char *acceptDataTyp)
{char file[256], name[ 128], *buff;
 INT16 hd; /* pipe-handle */
 LONG size;
 long reply_accept_type = 0;
 char ext[32];
 long ret=0;
 INT16 i;
 void *oldpipesig;
 /*..................... a priori pas de donnees reues ...*/
 *recept_type = 0;
 /*............ si pas de buffer de recopie donne .........*/
 /*             en entree le buffer de chargement          */
 /*             n'est pas detruit en sortie de fonction    */
 /*             et est retourne son adresse et il est      */
 /*             a la charge du programmeur de le liberer   */
 if (len_dst==0 || dst==0) len_dst=0;

 /*......... on signale le type de donnees acceptes ......*/
 /*          ici ARGS = liste d'arguments                  */
 strnset( ext, '\0', 32);
 memcpy(ext, acceptDataTyp, Min(strlen(acceptDataTyp),32));
 
 /*.............. on accepte le drag&drop .................*/
 hd = DragDropOpen( dd_ext, (ULONG*) ext, &oldpipesig);
 if (hd<=0)                                  return 0L;

 /*........... Lectures des infos reues.................. */
 /*            regarder si on peut recevoir les            */
 /*                       donnes                          */
 if (ddrtry( hd, name, file, ext, &size)<=0) return 0L;      /* erreur, */
 /*......... lecture dans notre liste si ..................*/
 /*          correspondance avec le type de  donnes       */
 /*          que nous envoie le pipe                       */
 for (i=0; i<8 && reply_accept_type==0; i++)
     {if (memcmp(acceptDataTyp + i*4, ext, 4)==0) 
         {memcpy(&reply_accept_type ,ext, 4);
          break;
         }
     }
 /*........SI il nous est envoye un type de donnes ................*/
 /*             non acceptes par notre liste d'entree              */
 /*             repondre on n'en veut pas                           */
 if (reply_accept_type==0) 
    {ddreply(hd, DD_NAK);
     ddclose( hd, oldpipesig);
                                             return 0L;
    }
 /* ......SINON repondre "on accepte les donnes" .............. */
 ddreply(hd, DD_OK);
 /*..... On lit les "size" octets des donnes dans le pipe .....*/
 buff = (char*)malloc(sizeof(char)*size+6);
 if (buff==0)   {ddclose( hd, oldpipesig);   return 0L;
                }
 /*......... lecture des donnes ...............*/
 if (Fread( hd, size, buff+4) == size)
    {buff[size+4] = 0;         /* fermer par zero de fin      */
     memcpy(buff, &size, 4);   /* placer longueur des donnes */
     /*...... Si il doit y avoir copie .........*/
     /*       des donnes dans la sortie        */
     /*       les copier                        */
     if (dst && len_dst>0) memcpy(dst, buff+4, len_dst);
    }
  /*..... si on n'a pas lu le nbr correct ......*/
  /*      erreur de lecture donc cassos         */
  else
     {ddclose( hd, oldpipesig);     /* fin du protocole */
      free( buff);                           return 0L;
     }
 /*....... si lecture buffer OK ................*/
 /*        sortir les infos                     */
 *recept_type  = reply_accept_type;  /* type de donnes recuperees      */
 if (len_dst==0)  {ret = (long)buff; /* retourner l'adresse du buffer   */
                  }
 else             {ret = size;       /* on sort longueur des donnes    */
                   free( buff);      /* a recuperer dans le drag & drop */
                  }
 ddclose( hd, oldpipesig); /* fin du protocole */
 return ret; 
}

/*--------------------------- DragDropCreate -----------------------------*/
/* ACTION: Routine pour l'envoyeur qui cree un pipe D&D                   */
/* ENTREE: WORD app_id: Iddentificateur d'application de l'expediteur     */
/*         WORD rcvr_id: Iddentificateur d'application du destinataire    */
/*         WORD window: handle GEM de la fenetre destinataire             */
/*         WORD mx, WORD my: coordonnes souris (lors lacher du D&D).     */
/*         WORD kbstate: etat touches controles du clavier(lors lacher    */
/*                       du D&D)                                          */
/*         ULONG format[8]; pointe sur un tableau de 8 longs qui recevra  */
/*                          les formats acceptes par le destinataire.     */
/*         void **oldpipesig: pointeur sur pointeur de void.              */
/* RETOUR: un nombre negatif ou = 0 si echec -1 echec -2 appl_write:      */
/*         le handle du pipe si tout est OK avec l'identificateur         */
/*                   ou extension dans le mot pointe par WORD *pipe       */

WORD DragDropCreate( WORD   app_id, WORD rcvr_id, WORD window, WORD mx, WORD my, 
                WORD kbstate, ULONG format[8], void **oldpipesig )
{BYTE   pipe[24];
 WORD   mbuf[8];
 LONG   handle_mask;
 WORD   handle, i;
/*............ placer nom du pipe............................*/
 strcpy( pipe, "U:\\PIPE\\DRAGDROP.AA" );
 pipe[18] = 'A' - 1;
/*............ tester un pipe libre .........................*/
 do{pipe[18]++;                                             /* Repeat last letter */
    if ( pipe[18] > 'Z' )                                   /* Not a letter of the alphabet? */
       {pipe[17]++;                                         /* Alter first letter of extension */
        if ( pipe[17] > 'Z' ) return( -1 );                 /* Could no pipe be opened? */
       }
    handle = (WORD) Fcreate( pipe, 0x02 );                  /* Create pipe, 0x02 means that EOF will be returned */
   } while ( handle == -36 ); /* #define EACCDN  -36   access denied */
 /*...... cassos si handle invalide ..........................*/
 if ( handle < 0 )   return( handle );                       /* Could the pipe not be created? */
 /*.......... envoyer message AP_DRAGDROP au destinataire ....*/
 mbuf[0] = AP_DRAGDROP;                                      /* Send Drap&Drop message  */
 mbuf[1] = app_id;                                           /* ID of own application */
 mbuf[2] = 0;
 mbuf[3] = window;                                           /* Window handle */
 mbuf[4] = mx;                                               /* x-coordinate of mouse */
 mbuf[5] = my;                                               /* y-coordinate of mouse */
 mbuf[6] = kbstate;                                          /* Keyboard status */
 mbuf[7] = (((WORD) pipe[17]) << 8 ) + pipe[18];             /* End of the pipe-name */
 if ( appl_write( rcvr_id, 16, mbuf ) == 0 )                 /* Error with appl_write()? */
    {Fclose( handle );    return( -2 );                      /* Close pipe */
    }
 /*............ attendre reponse destinataire ...............*/
 handle_mask = 1L << handle;
 i = Fselect( DD_TIMEOUT, &handle_mask, 0L, 0L );            /* Wait for reply */
 /*............. tester reponse .............................*/
 if ( i && handle_mask )                                     /* No timeout? */
    {BYTE   reply;
     if ( Fread( handle, 1L, &reply ) == 1 )                 /* Read reply from receiver */
        {if ( reply == DD_OK )                               /* All in order? */
            {if ( Fread( handle, DD_EXTSIZE, format ) == DD_EXTSIZE )   /* Read supported formats */
                {*oldpipesig = Psignal( SIGPIPE, (void *) SIG_IGN );   /* Unhook dispatcher */
                 return( handle );
                }
            }
        }
    }
 /*............. fermer le pipe si echec .....................*/
 Fclose( handle );                                           /* Close pipe */
 return( -1 );
}

/*--------------------------- ddcreate -----------------------------*/
/* ACTION: Routine pour l'envoyeur qui cree un pipe D&D             */
/* ENTREE: WORD *pipe: pointe sur un INT16 qui recevra: dans le poids */
/*                     fort et le faible l'extension du pipe cree.  */
/*                     Cette extention valide est automatiquement   */
/*                     recherche par la routine.                   */
/*         void **oldpipesig: pointeur sur pointeur de void.        */
/* RETOUR: un nombre negatif ou = 0 si echec:                       */
/*         le handle du pipe si tout est OK avec l'identificateur   */
/*                   ou extension dans le mot pointe par WORD *pipe */

WORD ddcreate(WORD *pipe, void **oldpipesig)
{LONG fd;
 Pipename[17] = 'A';
 Pipename[18] = 'A' - 1;
 /*............ recherche d'une extension valide ..........*/
 /*             et inocupe (pipe inocupe)                 */
 do
  {Pipename[18]++;
   if (Pipename[18] > 'Z')
      {Pipename[17]++;
       /*.... on a epuise toutes les letres ....*/
       /*     donc sortie sans avoir trouve     */
       /*     de pipe libre snif!               */
       if (Pipename[17] > 'Z')  break;
       else                     Pipename[18] = 'A';
      }
   /*.. FA_HIDDEN pour Pipe c'est comme ca! ....*/
   fd = Fcreate(Pipename, FA_HIDDEN);
  } while (fd == (LONG) -36); /* #define EACCDN  -36   access denied */
 /*......... cassos si pas de handle valide ....*/
    if (fd < 0L)                       return(-1);
 /*......... allez tout est OK on place ........*/
 /*          l'extension (identificateur)       */
 /*          du pipe dans le mot pointe         */
 *pipe = (Pipename[17] << 8) | Pipename[18];
 /*....... initialiser Psignal pour D&D ........*/
 *oldpipesig = Psignal( SIGPIPE, (void *) SIG_IGN ); /* Unhook dispatcher */
 return((WORD) fd);
}

/*--------------------- ddmessage -------------------------------*/
/* ACTION: envoie un message D&D: AP_DRAGDROP  une application  */
/* ENTREE: WORD apid: ID application destinataire du message     */
/*         WORD fd :  handle du Pipe c'est pas l'extension gaffe */
/*                    mais celui retourne par ddcreate           */
/*         WORD winid: handle GEM de la fenetre cible.           */
/*                     obtenu en faisant un wind_find()          */
/*         WORD mx, WORD my, WORD kstate: etat souris et touches */
/*                  de controles clavier.                        */
/*         WORD Pipename: extension du pipe (deux dernier octets */
/*                        de Pipename[]:  Pipename[17] & [18]    */
/*                        ou le mot: WORD *pipe initialise par   */
/*                        la fonction: ddcreate()                */
/*         void *oldpipesig: pointeur sur void.                  */
/* RETOUR: >0 pas d'erreur                                       */
/*        -1 DD_NAK: (non accept)  alors le pipe est ferme.     */
/*        -2 Timeout pas de reponse  alors le pipe est ferme.    */
/*        -3: erreur par appl_write()  alors le pipe est ferme.  */

WORD ddmessage(WORD apid, WORD fd, WORD winid, WORD mx, WORD my, WORD kstate, 
               WORD Pipename, void *oldpipesig)
{BYTE c;
 WORD i, msg[8];
 LONG fd_mask;
 /*.............. AES-Message definition et envoi ............. */
 msg[0] = AP_DRAGDROP;
 msg[1] = Appl_id;
 msg[2] = 0;
 msg[3] = winid;
 msg[4] = mx;
 msg[5] = my;
 msg[6] = kstate;
 msg[7] = Pipename;
 i      = appl_write(apid, 16, msg);
 /*............ si echec envoi cassos .........................*/
 if (i == 0)
    {ddclose(fd, oldpipesig);
     return(-3);
    }
 /*........... analyse reaction du destinataire ...............*/
 fd_mask = (1L << fd);
 
 /*............ ouvrir le pipe ................................*/
 i = Fselect(DD_TIMEOUT, &fd_mask, 0L, 0L);
 /*........... si Timeout cassos ..............................*/
 if (!i || !fd_mask)
    {ddclose(fd, oldpipesig);
     return(-2);
    }
 /*........... le destinataire refuse cassos...................*/
 /*      lire un caractere qui doit etre  0=DD_OK (1=DD_NAK)   */
 if (Fread(fd, 1L, &c) != 1L)
    {ddclose(fd, oldpipesig);
     return(-1);         /* cassos si on n'a pas pu lire 1 car */
    }
 if (c != DD_OK)
    {ddclose(fd, oldpipesig);
     return(-1);         /* cassos si ce car = (1=DD_NAK)      */
    }
 /*............ tout est OK on continue .......................*/
 return(1);
}

/*------------------- ddrexts -----------------------------------*/
/* ACTION: Demande au destinataire la liste des types de donnees */
/*         qu'il peut accepter. liste de 32 long ou byte[4]      */
/* ENTREE: WORD fd :  handle du Pipe c'est pas l'extension gaffe */
/*                    mais celui retourne par ddcreate           */
/*         BYTE *exts: pointe sur zone de 32 octets ou sera      */
/*                     retourne les extensions admises.         */
/*         void *oldpipesig: pointeur sur  void.                 */
/* RETOUR: WORD :  >0 pas d'erreur.                              */
/*                <=0 0 alors le pipe est ferme.                 */

WORD ddrexts(WORD fd, BYTE *exts,  void *oldpipesig)
{if (Fread(fd, DD_EXTSIZE, exts) != DD_EXTSIZE)
    {ddclose(fd, oldpipesig);
     return(-1);
    }
 return(1);
}

/*------------------------- ddsexts --------------------------------------*/
/* ACTION: envoie a l'expediteur la liste des types de donnes acceptes  */
/*         cette liste est donc de 32 octets, chaque type etant 4 octets  */
/*         du type "ARGS" ".IMG" etc..  elle peut accepter 8 types de     */
/*         donnes. Si elle n'est pas complete (exemple seulement deux    */
/*         types de donnes) le reste doit etre rempli de zeros.          */
/* ENTREE: BYTE *data_typ: pointe sur liste type de donnees acceptes     */
/*                         "ARGS" "PATH" ".IMG" ".TXT" etc...             */
/*         void *oldpipesig: pointeur sur  void.                          */
/* RETOUR:  >0 tout est OK                                                */
/*        <= 0 erreur d'ecriture (le pipe est alors ferme)                */

WORD ddsexts(WORD fd, BYTE *data_typ, void *oldpipesig)
{if (Fwrite(fd, DD_EXTSIZE, data_typ) != DD_EXTSIZE)
    {ddclose(fd, oldpipesig);
     return(-1);
    }
 return(1);
}


/*------------------- ddstry ---------------------------------------*/
/* ACTION: demande au destinataire si il accepte un certain         */
/*         type de donnees dont il presente les caracteristiques    */
/*         data_typ, LONG size etc ....                             */
/* ENTREE: WORD fd :  handle du Pipe c'est pas l'extension gaffe    */
/*                    mais celui retourne par ddcreate              */
/*         BYTE *data_typ: pointe sur type de donnees a envoyer     */
/*                         "ARGS" "PATH" ".IMG" ".TXT" etc...       */
/*              BYTE *text: descripteur des donnes exemple:        */
/*                       "Caractere ASCII" est optionnel on peut    */
/*                        passer une chaine vide.                   */
/*              BYTE *name: non de fichier associe  la donne      */
/*                        ce parametre est aussi optionnel et       */
/*                        peut etre une chaine vide.                */
/*              LONG size: longueur des donnes  transmettre       */
/* RETOUR: un INT16:                                                  */
/*       DD_OK  le destinataire ou cible accepte les donnes        */
/*       DD_NAK le D&D doit etre interrompu                         */
/*       DD_EXT le cible refuse les donnes telles quelles          */
/*       DD_LEN le cible ne peut accepter la longueur               */
/*     DD_TRASH le cible repond que donnes sont sur une poubelle   */
/*   DD_PRINTER le cible repond que donnes sont sur une Imprimante */
/* DD_CLIPBOARD le cible repond que donnes sont sur le CLIPBOARD   */

WORD ddstry(WORD fd, BYTE *data_typ, BYTE *text, BYTE *name, LONG size)
{BYTE c;
 WORD hdrlen, i;
 /*................. calcul longueur header .........................*/
 /*              4 Bytes fr "ext", 4 Bytes fr "size",              */
 /*              2 Bytes fr Stringendnullen                         */
 hdrlen = (WORD) (4 + 4 + strlen(text)+1 + strlen(name)+1);
 /*.......Si echec de l'ecriture de la longueur du header cassos ....*/
 if (Fwrite(fd, 2L, &hdrlen) != 2L)             return(DD_NAK);
 /*.......Si echec de l'ecriture du header lui meme  cassos .........*/
    i  = (WORD) Fwrite(fd, 4L, data_typ);
    i += (WORD) Fwrite(fd, 4L, &size);
    i += (WORD) Fwrite(fd, strlen(text)+1, text);
    i += (WORD) Fwrite(fd, strlen(name)+1, name);
 /*.... si longueur retournee pas egale  longueur envoye cassos ...*/
 if (i != hdrlen)                               return(DD_NAK);
 /*.......... si echec de la lecture de la reponse cassos ...........*/
 if (Fread(fd, 1L, &c) != 1L)                   return(DD_NAK);
 /*.......... retourner la reponse qui peut etre aussi DD_NAK .......*/
 /*                          NIAK NIAK NIAK !!!                      */
 return(c);
}


/*-------------------------- ddclose --------------------------------*/
/* ACTION: ferme le pipe et met fin aux procedures d'echanges        */
/*         peut etre utilisee par l'expediteur et le destinataire.   */
/* ENTREE: WORD fd :  handle du Pipe c'est pas l'extension gaffe     */
/*                    mais celui retourne par ddcreate               */
/*         void *oldpipesig:  pointeur de void.                      */
VOID ddclose(WORD fd, void *oldpipesig )
{/*......... replacer ancien handle pour Psignal ....................*/
 Psignal( SIGPIPE, oldpipesig );   /* Enter old dispatcher once more */
 Fclose(fd);
}

/*--------------------- DragDropOpen ----------------------------------*/
/* ACTION: ouvre un pipe deja cree par un expediteur, en reponse au    */
/*         au message AP_DRAGDROP.                                     */
/* ENTREE: WORD ext_name: contient en poids fort et poids faible les   */
/*                        deux caracteres d'extension du pipe          */
/*                    reu par le destinataire lors reception du       */
/*                    du message: AP_DRAGDROP, dans le mot N7:        */
/*         ULONG format[8]: pointeur sur les formats supportes         */
/*         void **oldpipesig: pointeur sur pointeur de void.           */
/* RETOUR:  >0 handle du pipe.                                         */
/*         <=0 erreur                                                  */

WORD DragDropOpen( WORD ext_name, ULONG format[8], void **oldpipesig )
{WORD   handle;
 BYTE   reply;
 BYTE   pipe_name[24];
 strcpy( pipe_name, "U:\\pipe\\dragdrop.AA" );
 pipe_name[17] = (ext_name & 0xff00) >> 8;
 pipe_name[18] =  ext_name & 0x00ff;
 handle = (WORD) Fopen( pipe_name, FO_RW );                /* Inquire pipe handle */
 if ( handle < 0 )                        return( -1 );
 reply = DD_OK;                                            /* Program supports Drag & Drop */
 *oldpipesig = Psignal( SIGPIPE, (void *) SIG_IGN );       /* Ignore signal */
 if ( Fwrite( handle, 1, &reply ) == 1 )
    {if ( Fwrite( handle, DD_EXTSIZE, format ) == DD_EXTSIZE )
         return( handle );
    }
 ddclose( handle, *oldpipesig );                           /* Close pipe */
 return( -1 );
}


/*--------------------- ddopen ----------------------------------------*/
/* ACTION: ouvre un pipe deja cree par un expediteur, en reponse au    */
/*         au message AP_DRAGDROP.                                     */
/* ENTREE: WORD ddnam: extension du pipe  ouvrir tel qu'il a ete recu */
/*                     par le destinataire lors reception du message:  */
/*                     AP_DRAGDROP, dans le mot N7: exemple pipe[7]   */
/*         BYTE ddmsg: DD_OK on accepte   DD_NAK on refuse             */
/*         void **oldpipesig: pointeur sur pointeur de void.           */
/* RETOUR:  >0 handle du pipe.                                         */
/*         <=0 erreur                                                  */
WORD ddopen(WORD ddnam, BYTE ddmsg, void **oldpipesig)
{LONG fd;
 ddmsg = DD_OK;
 Pipename[17] = (ddnam & 0xff00) >> 8;
 Pipename[18] = ddnam & 0x00ff;
 fd           = Fopen(Pipename, FO_RW);
 if (fd < 0L)  return(-1);
 /*....... initialiser Psignal pour D&D ........*/
 *oldpipesig = Psignal( SIGPIPE, (void *) SIG_IGN );      /* Ignore signal */
 /*....... ecriture de la reponse ..............*/
 if (Fwrite((WORD) fd, DD_EXTSIZE, &ddmsg) == DD_EXTSIZE) return((WORD) fd);
 
 ddclose((WORD) fd, *oldpipesig);
 return(-1);           /* si echec de l'ecriture reponse cassos */
 
}

/*------------------------- ddrtry -----------------------------------*/
/* ACTION: demande  l'expediteur, en lui precisant le type de donnee */
/*         que l'on accepte de recevoir, des renseignements divers    */
/*         dessus notamment la longueur.                              */
/* ENTREE: WORD fd :  handle du Pipe c'est pas l'extension gaffe      */
/*                    mais celui retourne par ddcreate                */
/*         BYTE *whichext: pointe sur type de donnees que l'on        */
/*                         souhaite recevoir. zone de 5 octets.       */
/*                         "ARGS" "PATH" ".IMG" ".TXT" etc...         */
/* SORTIE:                                                            */
/*         BYTE *whichext: est rempli apre appel avec type de donnees */
/*                         reues "ARGS" "PATH" ".IMG" ".TXT" etc...  */
/*                         si different de celui en entree alors      */
/*                         erreur.                                    */
/*         BYTE *text: descripteur des donnes exemple:               */
/*                    "Caractere ASCII" initialisee par la fonction   */
/*                    la longueur de la zone pointee par text doit    */
/*                    etre au minimum de: DD_NAMEMAX 128 octets       */
/*         BYTE *name: non de fichier associe  la donne             */
/*                     ce parametre est initialise par la fonction    */
/*                     la longueur de la zone pointee par name doit   */
/*                     etre au minimum de: DD_NAMEMAX 128 octets      */
/*         LONG *size: pointe sur longueur des donnes  ecevoir      */
/*                     cette valeur est initialise en retour.        */
/* RETOUR: >0 Tout est OK                                             */
/*         -1 erreur                                                  */

WORD ddrtry(WORD fd, BYTE *name, BYTE *file, BYTE *whichext, LONG *size)
{BYTE buf[DD_NAMEMAX * 2];
 WORD hdrlen, i, len;
 /*................... lire les deux octets longueur header ..........*/
 if (Fread(fd, 2L, &hdrlen) != 2L)                        return(-1);
 /*...... si longueur header lue inferieure au minimum erreur ........*/
 /*               il reste au minimum 11 - 2 = 9 octets a lire        */
 if (hdrlen < 9)  /* erreur taille incorrecte */          return(-1);    
 /*............ lecture de l'extension ...............................*/
 if (Fread(fd, 4L, whichext) != 4L)                       return(-1);
 /*............ lecture de la taille des donnes .....................*/
 if (Fread(fd, 4L, size) != 4L)                           return(-1);
 hdrlen -= 8;    /* on a lu 8 octets  */
 /*............. lecture des chaines nom et file .....................*/
 /*               ce stade la longueur restante est celle des        */
 /*              deux chaines  lire.                                 */
 /*...... limiter lecture restante au mximun buffer possible .........*/
 if (hdrlen > DD_NAMEMAX*2) i = DD_NAMEMAX*2; 
 else                       i = hdrlen;
 len = i;
 /*............. lire le reste des donnes ...........................*/
 if (Fread(fd, (LONG) i, buf) != (LONG) i)                return(-1);
 /*............. separer les deux chaines ............................*/
 hdrlen -= i;  /* il se peut qu'il y ait des octets en plus  lire ?? */
 strncpy(name, buf, DD_NAMEMAX); /* copier en sortie la 1ere chaine   */
 i = (WORD) strlen(name) + 1;    /* pointer apres 1er chaine          */
 if (len - i > 0)                /* si il reste donnes les lire      */
     strncpy(file, buf + i, DD_NAMEMAX);
 else
     file[0] = '\0';
 /*............ epuiser le reste des octets au cas ou il en reste ....*/
 while (hdrlen > DD_NAMEMAX*2)
    {if (Fread(fd, DD_NAMEMAX*2, buf) != DD_NAMEMAX*2)    return(-1);
     hdrlen -= DD_NAMEMAX*2;
    }
 if (hdrlen > 0)
    {if (Fread(fd, (LONG) hdrlen, buf) != (LONG) hdrlen)  return(-1);
    }
 return(1);
}

/*-------------------- ddreply ----------------------------------------*/
/* ACTION: envoie 1 octet exemple: DD_OK ou DD_NAK au pipe.            */
/* ENTREE: WORD fd :  handle du Pipe c'est pas l'extension gaffe       */
/*                    mais celui retourne par ddcreate ou ddopen       */
/*         BYTE ack: octet  envoyer                                   */
/* RETOUR: >=0 Tout est OK                                             */
/*           0 erreur                                                   */

WORD ddreply(WORD fd, BYTE ack)
{if (Fwrite(fd, 1L, &ack) != 1L)
    {Fclose( fd );        /* Close pipe */
     return(0);
    }
 return(1);
}

/*------------------------- Parse_ARGS --------------------------------*/
/* ACTION: transforme une liste d'arguments soit separes par des       */
/*         espaces, en une suite de chaines terminees par des zeros de */
/*         fin. Si les arguments sont entre quotes: exemple            */
/*         'arguments1' 'argument2'  etc... cela marche aussi et les   */
/*         espaces ne seront pas transformes en zero de fin.           */
/* ENTREE: char *str: chaine d'arguments termine par un zero.         */
/* RETOUR: nombre d'arguments.                                         */

INT16  Parse_ARGS(char *str)
{INT16     cnt    = 1;
 char    *c     = str;
 INT16   in_quote = FALSE;

 while (*c)
   {switch (*c)
      {case ' ' :
            if (!in_quote)
               {*c = '\0';
                cnt++;
               }
            break;
       case '\'' :
            strcpy(c, c + 1);
            if (!in_quote) in_quote = TRUE;
            else
              {if (*c != '\'')
                  {in_quote = FALSE;
                   *c       = 0;
                   if (c[1]) cnt++;
                  }
              }
            break;
       default:
       break;
      } /* switch */
    c += 1;
   } /* while */
 return cnt;
}

