
#ifndef __MEM_MNGR__
#define __MEM_MNGR__

/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*\\\\\\\\\\\\\\\\\\\\\\\\\\ MEM_MNGR.H  \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*\\\\\\\\\\\\\\\\\\\ par Roland SEVIN pour EZ_GEM \\\\\\\\\\\\\\\\\\\\\\\\*/
/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*              Fichier d'entete  inclure dans tout source desirant       */
/*              profiter des fonctionnalits de MEM_MNGR.C                 */
/*              gestion des erreurs d'allocation desallocation             */
/*              et corruption de la memoire.                               */
/*                                                                         */
/* AUTEUR: Roland SEVIN. Utilise en partie une methode de double chainage  */
/*                       publie sur:  fr.comp.sys.atari  par notre grand  */
/*                                     gourou Jean LUSETTI merci  lui !!! */
/*                                     au passage le code provient des     */
/*                                     sources de NT, mais fallait trouver */
/*              Toutes les fonctions sont largement commentes et je vous  */
/*              invite  lire les commentaires.                            */
/*              Ce source est libre de droit, cependant si des corrections */
/*              et ameliorations y sont apportes par vos soins, ayez la   */
/*              gentillesse de m'en faire part: upsis@club-internet.fr     */
/*                                                                         */
/* MODE D'EMPLOI:                                                          */
/*              1) incluez dans votre projet le source: MEM_MNGR.C         */
/*              2) placez la mention:   #include "MEM_MNGR.H"  en tete de  */
/*                 tous vos fichiers .C (sauf celui de MEM_MNGR.C)         */
/*                 ou placez le dans votre .H general si vous en avez un.  */
/*                 Bref il faut que tout source .C utilisant les fonctions */
/*                 malloc(), calloc(), realloc(), free(), inclue le        */
/*                 fichier: MEM_MNGR.H pour que cela marche !!!.           */
/*                 vous pouvez tres bien faire une version debug style:    */
/*                                                                         */
/*                 #if __DEBUG__                                           */
/*                    #include "MEM_MNGR.H"                                */
/*                 #endif                                                  */
/*                                                                         */
/*                 bon vous etes grand et faites un peu ce que vous voulez */
/*              3) -> Juste apres le main() faites appel  la fonction     */
/*                    d'initialisation:  InitMemMngr(int mode);            */
/*                 -> Donnez le chemin et nom du fichier ou vous desirez   */
/*                    que les erreurs soient reportes avec la fonction:   */
/*                                       SetErreurName();                  */
/*                 -> Juste avant la sortie du programme, faites appel    */
/*                    la fonction de nettoyage des blocs non librs:      */
/*                                       FreeAllMemory();                  */
/*              4) Exemple:                                                */
/*                                                                         */
/*     #include "MEM_MNGR.H"                                               */
/*                                                                         */
/*     int main ( int argc, const char *argv[], const char *envp[] )       */
/*        {                                                                */
/*         .......... initialiser Le Memory Manager .................      */
/*         InitMemMngr( TO_FILE | TO_SCREEN | TO_EZ_EDIT );                */
/*                                                                         */
/*         ......... definir le fichier de sortie des erreurs .......      */
/*         SetErreurName("E:\\MY_PRJ\\", "MEM_ERR.ERR", ERR_ERASE);        */
/*                                                                         */
/*         ......... envoyer un petit message dans la ...............      */
/*                 fenetre LOGOUT d'EZ_EDIT si elle est ouverte.           */
/*                 Si vous cliquez sur ce message, EZ_EDIT                 */
/*                 place son curseur  l'endroit ou il a ete               */
/*                 genere dans le source !! utile pour certains            */
/*                 debug.                                                  */
/*                                                                         */
/*         SetMemMngrMode(TO_EZ_EDIT);                                     */
/*         OutMessage("Salut la compagnie, me voici me voila");            */
/*         SetMemMngrMode(TO_FILE | TO_SCREEN | TO_EZ_EDIT);               */
/*                     .                                                   */
/*                     .                                                   */
/*                     .                                                   */
/*         .......... liberer tous les blocs non desalous ..........      */
/*                   (les signale dans le fichier d'erreurs)               */
/*         FreeAllMemory();                                                */
/*         exit(0);                                                        */
/*        }                                                                */
/*                                                                         */
/*             5) A tout moment il vous est possible de verifier si une    */
/*                adresse memoire est corrompue en faisant appel :        */
/*                                                                         */
/*                int VerifyBlk(void *test_ptr, int mode);     qui verifie */
/*                    si une adresse memoire a ete corrompue.              */
/*                                                                         */
/*                int VerifyAllBlk(int mode);                  qui verifie */
/*                    tous les blocs memoire et signale ceux corrompus     */
/*                    dans le fichier d'erreur.                            */
/*                                                                         */
/* Les fonctions de reservation qui vous changeront la vie sont celles     */
/* qui permettent un typage de la memoire lors de la reservation.          */
/*                                                                         */
/*   malloc_ID()                                                           */
/*   realloc_ID()                                                          */
/*   calloc_ID()                                                           */
/*   FreeAll_ID()                                                          */
/*                                                                         */
/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*\\\\\\\\\\\\\\\\\\\\\\\\\\ D E F I N E S  \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*.................... modes d'affichage des erreurs ......................*/

#define ERR_ERASE   1           /* parametre:  fichier erreur effac       */
#define ERR_APPEND  0           /* parametre:  fichier erreur non effac   */

#define TO_FILE    1             /* parametre:  sortie  erreur sur fichier */
#define TO_SCREEN  2             /* parametre:  sortie erreur sur ecran    */
#define TO_EZ_EDIT 4             /* parametre:  sortie erreur dans EZ_EDIT */


/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
/*\\\\\\\\\\\\\\\ P R O T O T Y P E S  F O N C T I O N S \\\\\\\\\\\\\\\\\\*/
/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

/*......................... gestion de la memoire .........................*/
/*                 les fonctions suivantes sont celles                     */
/*                 qui vont remplacer les fonctions standard               */
/*                 de gestion de la memoire. vous n'avez pas              */
/*                 y faire appel directement.                              */

extern void* MyMalloc( size_t size , long lign, char *file, long Id);
extern void* MyRealloc( void *ptr, size_t size, long lign , char *file, long Id);
extern void* MyCalloc( size_t num_elems, size_t elem_size , long lign, char *name, long Id);
extern void  MyFree( void *ptr , long lign , char *file);

/*................ pour les quatres fonctions suivantes il ................*/
/*                 vaut mieux appeler leurs equivalentes                   */
/*                 representes par les macro qui vous                     */
/*                 dispensent d'ajouter __LINE__ et __FILE__               */
/*                 lors de l'appel. (la macro le fait pour vous)           */

/*...... APPEL EQUIVALENT: VerifyBlk(void *test_blk, int mode) ............*/
/*                         definie plus loin par une macro                 */
extern int   BlkCheck( void *test_blk, int mode, long lign, char *name);

/*...... APPEL EQUIVALENT: VerifyAllBlk(int mode) .........................*/
/*                         definie plus loin par une macro                 */
extern void  MemoryCheck(int mode, long lign, char *name);

/*...... APPEL EQUIVALENT: FreeAllMemory() ................................*/
/*                         definie plus loin par une macro                 */
extern void  ExitMemMngr(long lign, char *name);

/*...... APPEL EQUIVALENT: FreeAll_ID() ...................................*/
/*                         definie plus loin par une macro                 */
extern long  FreeMem_ID(long lign, char *name, long iD);

/*----------------------------- InitMemMngr ---------------------------------*/
/* ACTION: fonction d'initialisation generale permettant d'initialiser       */
/*         les pointeurs du premier bloc de base. Cette fonction est        */
/*         appeler en premier dans main().                                   */
/* ENTREE: int mode: mode de sortie par defaut des erreurs.                  */
/*                   TO_FILE  l'erreur est reporte dans le fichier d'erreur */
/*                   TO_SCREEN l'erreur est affiche  l'ecran               */
/*                   TO_EZ_EDIT l'erreur est envoye  EZ_EDIT               */
/*       ces trois modes peuvent etre combins (TO_FILE|TO_SCREEN|TO_EZ_EDIT)*/
/*       et modifis par la fonction: int SetMode(int mode)                  */
/* SORTIE: neant                                                             */

extern void  InitMemMngr(int mode);

/*-------------------------- SetMode ----------------------------------------*/
/* ACTION: met en place le mode par defaut de sortie des erreurs.            */
/* ENTREE: int mode: TO_FILE  l'erreur est reporte dans le fichier d'erreur */
/*                   TO_SCREEN l'erreur est affiche  l'ecran               */
/*                   TO_EZ_EDIT l'erreur est envoye  EZ_EDIT               */
/*       ces trois modes peuvent etre combins (TO_FILE|TO_SCREEN|TO_EZ_EDIT)*/
/*       et modifis par la fonction: int SetMode(int mode)                  */
/* SORTIE: int: qui est l'ancien mode                                        */

extern int   SetMode(int mode);

/*---------------------- SetErreurName ----------------------------*/
/* ACTION: permet de donner le chemin et le nom du fichier         */
/*         d'erreurs par defaut.                                   */
/* ENTREE: char *path, char *nom: pointe sur le chemin et le nom   */
/*                                du fichier d'erreurs. Si nom = 0 */
/*                                alors path doit etre complet     */
/*                                sinon nom est concatene  path   */
/*         int mode: 1/ si il existe un fichier d'erreur portant   */
/*                      le nom de celui donn en entre alors il   */
/*                      est effac.                                */
/*                   0/ si il existe un fichier d'erreur portant   */
/*                      le nom de celui donn en entre alors il   */
/*                      n'est pas effac (les donnes seront alors */
/*                                        rajoutes)               */
/* SORTIE: neant.                                                  */

extern void  SetErreurName(char *path, char *nom, int mode);

/*-------------------------------- OutErreur --------------------------------*/
/* ACTION: ajoute la chaine: char *err dans le fichier d'erreurs             */
/*         par defaut et ou l'affiche selon mode de sortie par defaut.       */
/*         Je vous conseille d'utiliser: void OutMessage(char *err) qui vous */
/*         dispense de fournir: __LINE__ et __FILE__                         */
/* ENTREE: char *err: pointe sur la chaine  rajouter au fichier             */
/*         d'erreurs par defaut et ou  afficher.                            */
/*         long lign: ligne du source ou est appele la fonction (mettre     */
/*                    __LINE__   )                                           */
/*         char *name: nom du source ou est appele la fonction. (mettre     */
/*                    __FILE__                                               */
/* SORTIE: neant.                                                            */
/* NOTE:   La chaine:  char *err est ajoute dans le fichier d'erreurs et ou */
/*         affiche  l'ecran selon le mode mis en place par la fonction:    */
/*                       void InitMemMngr(int mode)  o   int mode           */
/*          = TO_FILE  l'erreur est reporte dans le fichier d'erreur        */
/*          = TO_SCREEN l'erreur est affiche  l'ecran                      */
/*          = TO_EZ_EDIT l'erreur est envoye  EZ_EDIT ya plus qu'a cliquer!*/
/*       ces trois modes peuvent etre combins (TO_FILE|TO_SCREEN|TO_EZ_EDIT)*/
/*       et modifis par la fonction: int SetMode(int mode)                  */

extern void OutErreur(char *err, long lign, char *name);

/*-------------------------------- OutMessage -------------------------------*/
/* ACTION: ajoute la chaine: char *err dans le fichier d'erreurs             */
/*         par defaut et ou l'affiche selon mode de sortie par defaut        */
/*         cette fonction permet d'ajouter au fichier d'erreur un message    */
/*         de son choix. cela peut etre utile par exemple si l'on veut faire */
/*         un fichier .LOG ou journal etc....                                */
/* ENTREE: char *err: pointe sur la chaine  rajouter au fichier             */
/*         d'erreurs par defaut et ou  afficher.                            */
/* SORTIE: neant.                                                            */
/* NOTE:   La chaine:  char *err est ajoute dans le fichier d'erreurs et ou */
/*         affiche  l'ecran selon le mode mis en place par la fonction:    */
/*                       void InitMemMngr(int mode)  o   int mode           */
/*          = TO_FILE  l'erreur est reporte dans le fichier d'erreur        */
/*          = TO_SCREEN l'erreur est affiche  l'ecran                      */
/*          = TO_EZ_EDIT l'erreur est envoye  EZ_EDIT ya plus qu'a cliquer!*/
/*       ces trois modes peuvent etre combins (TO_FILE|TO_SCREEN|TO_EZ_EDIT)*/
/*       et modifis par la fonction: int SetMode(int mode)                  */
/*                                                                           */
/* PROTO: void OutMessage(char *err)     defini par la macro suivante:       */

#define OutMessage(a)      OutErreur((a),  , __LINE__, __FILE__)


/*----------------------- VerifyBlk -----------------------------------------*/
/* ACTION: verifie si l'adresse memoire donne en entree est corrompue.      */
/*         cette fonction est utile afin de savoir si apres appel  une      */
/*         fonction, celle ci abime une zone memoire sur laquelle elle       */
/*         doit travailler.                                                  */
/* ENTREE: void *test_ptr: pointeur initialise par une des fonctions         */
/*                         malloc, calloc, realloc, dont il faut verifier    */
/*                         l'integrite de la zone memoire pointe.           */
/*         int mode: TO_FILE  l'erreur est reporte dans le fichier d'erreur */
/*                   TO_SCREEN l'erreur est affiche  l'ecran               */
/*                   TO_EZ_EDIT l'erreur est envoye  EZ_EDIT               */
/*       ces trois modes peuvent etre combins (TO_FILE|TO_SCREEN|TO_EZ_EDIT)*/
/*       et modifis par la fonction: int SetMode(int mode)                  */
/* SORTIE: 1/ une erreur a ete trouve.                                      */
/*         0/ pas d'erreur trouve.                                          */
/* PROTO:  int VerifyBlk(void *test_ptr, int mode);                          */
/*                                             defini par la macro suivante: */

#define VerifyBlk(a,b)     BlkCheck((a), (b) , __LINE__, __FILE__)

/*-------------------- VerifyAllBlk -----------------------------------------*/
/* ACTION: scanne tous les blocs allous au moment de l'appel et verifie si  */
/*         ils sont corrompus et ecrit dans le fichier d'erreur si c'est le  */
/*         cas.                                                              */
/* ENTREE: int mode: TO_FILE  l'erreur est reporte dans le fichier d'erreur */
/*                   TO_SCREEN l'erreur est affiche  l'ecran               */
/*                   TO_EZ_EDIT l'erreur est envoye  EZ_EDIT               */
/*       ces trois modes peuvent etre combins (TO_FILE|TO_SCREEN|TO_EZ_EDIT)*/
/*       et modifis par la fonction: int SetMode(int mode)                  */
/*                                                                           */
/* SORTIE: neant.                                                            */
/* PROTO: void VerifyAllBlk(int mode);  defini par la macro suivante:        */

#define VerifyAllBlk(a)    MemoryCheck((a), __LINE__, __FILE__)


/*-------------------- FreeAllMemory ----------------------------------------*/
/* ACTION: libere tous les blocs alloues de la liste chaine. Cette fonction */
/*          appeler en fin de programme permet d'etre certain que toute la  */
/*         memoire sera libre en fin de programme. Normalement il ne       */
/*         devrait  pas y avoir de bloc restant  desallouer...              */
/* ENTREE: neant.                                                            */
/* SORTIE: Neant.                                                            */
/*                                                                           */
/* NOTE:   si des erreurs sont trouves lors de la desallocation, elles sont */
/*         reportes dans le fichier d'erreurs et ou affiches  l'ecran     */
/*         selon le mode mis en place par la fonction:                       */
/*                       void InitMemMngr(int mode)  o   int mode           */
/*          = TO_FILE  l'erreur est reporte dans le fichier d'erreur        */
/*          = TO_SCREEN l'erreur est affiche  l'ecran                      */
/*          = TO_EZ_EDIT l'erreur est envoye  EZ_EDIT ya plus qu'a cliquer!*/
/*       ces trois modes peuvent etre combins (TO_FILE|TO_SCREEN|TO_EZ_EDIT)*/
/*       et modifis par la fonction: int SetMode(int mode)                  */
/*                                                                           */
/* PROTO: void FreeAllMemory();  defini par la macro suivante:               */

#define FreeAllMemory()    ExitMemMngr(__LINE__, __FILE__)

/*-------------- malloc_ID ---- calloc_ID ------- realloc_ID ----------------*/
/*ACTION: comme malloc, calloc, realloc  sauf que vous specifiez en plus un  */
/*        parametre d'identificateur de votre choix, permettant lorsque      */
/*        vous appelerez par la suite la  fonction:                          */
/*                free_all_ID() de ne liberer que les blocs possedant cet    */
/*                              identificateur.                              */
/*ENTREE: size_t size: taille du bloc  reserver.                            */
/*        long Id:     identificateur de ce bloc. (selon votre imagination)  */
/*        size_t n_memb: nombre de bloc  reserver (pour calloc_ID)          */
/*        sortie void *: pointeur sur la zone reserve ou zero si echec.     */
/*                                                                           */
/* PROTO: void* malloc_ID(  size_t size, long iD);                           */
/* PROTO: void* realloc_ID( size_t size, long iD);                           */
/* PROTO: void* calloc_ID(  size_t n_memb, size_t size, long iD);            */
/*                                                                           */
/* ATTENTION: L'identificateur iD = 0 est reserve par les fonctions:         */
/*            malloc(), calloc(), realloc() redefinies                       */

#define  malloc_ID(a,b)      MyMalloc((a), __LINE__, __FILE__ , (b))
#define  realloc_ID(a,b,c)   MyRealloc((a),(b), __LINE__, __FILE__ , (c))
#define  calloc_ID(a,b,c)    MyCalloc((a),(b), __LINE__, __FILE__ , (c))

/*-------------------- FreeAll_ID -------------------------------------------*/
/* ACTION: libere tous les blocs alloues de la liste chaine, dont l'identi- */
/*         ficateur est commun.                                              */
/*         long Id: identificateur des blocs  liberer.                      */
/* SORTIE: long: nombre de blocs liberes.                                    */
/*                                                                           */
/* NOTE:   si des erreurs sont trouves lors de la desallocation, elles sont */
/*         reportes dans le fichier d'erreurs et ou affiches  l'ecran     */
/*         selon le mode mis en place par la fonction:                       */
/*                       void InitMemMngr(int mode)  o   int mode           */
/*          = TO_FILE  l'erreur est reporte dans le fichier d'erreur        */
/*          = TO_SCREEN l'erreur est affiche  l'ecran                      */
/*          = TO_EZ_EDIT l'erreur est envoye  EZ_EDIT ya plus qu'a cliquer!*/
/*       ces trois modes peuvent etre combins (TO_FILE|TO_SCREEN|TO_EZ_EDIT)*/
/*       et modifis par la fonction: int SetMode(int mode)                  */
/* PROTO: long  FreeAll_ID(long iD);                                         */

#define FreeAll_ID(a)      FreeMem_ID(__LINE__, __FILE__, (a))

/*...................... detournement des fonctions standard ................*/
/*                       de gestion de la memoire                            */
#ifndef __EZMEM_INIT__ /* ce flag n'est defini que dans: MemMngr.C ou bien   */
                       /* sur il ne faut surtout pas redefinir les fonctions */
                       /* de gestion de la memoire !!!                       */
  #undef  malloc(a)
  #define malloc(a)          MyMalloc((a), __LINE__, __FILE__ , 0L)

  #undef  calloc(a,b)
  #define calloc(a,b)        MyCalloc((a),(b), __LINE__, __FILE__ , 0L)

  #undef  free(a)
  #define free(a)            MyFree((a), __LINE__, __FILE__)

  #undef  realloc(a,b)
  #define realloc(a,b)       MyRealloc((a),(b), __LINE__, __FILE__ , 0L)
#endif                 /* __EZMEM_INIT__ */
/*..........................................................................*/
#endif